#ifndef MUFS_H
#define MUFS_H

#include <stdint.h>
#include <stddef.h>

/* Chybove kody */
#define EOK		(0)
#define ENO_INODES_LEFT	(-1)
#define ENO_SPACE_LEFT	(-2)
#define EINVALID_LINK	(-3)
#define EIO_ERROR	(-4)
#define EINVALID_PATH	(-5)
#define EFILE_NOT_FOUND	(-6)
#define EFILE_EXISTS 	(-7)
#define ENO_FDS         (-8)

/* velikosti a vlastnosti struktur */
#define INODE_SIZE	  (offsetof(struct inode, id))  /* velikost perzistentni casti inode v bytech */
#define SUPER_BLK_SIZE    (1024)                        /* velikost superbloku v bytech */
#define DIRECT_LINKS	  (6)	                        /* pocet primych bloku v inode */
#define EMPTY_CLUSTER     (INT32_MAX)                   /* indikator prazdneho clusteru */
#define MAX_OPEN_FILES    (16)                          /* maximalni pocet otevrenych souboru */
#define INODE_TABLE_SIZE  (MAX_OPEN_FILES * 2)          /* velikost hash-tabulky s otevrenymi soubory */
#define MAX_FILENAME_SIZE (256)                         /* maximalni delka cesty + \0 */

/* pomocna makra */


/* pocet neprimych odkazu v jednom bloku */
#define INDIRECT_LINKS(fs) (((fs)->blk_size * (fs)->blks_per_cluster) / sizeof(cluster_id))

/* velikost cluster v bytech */
#define CLUSTER_SIZE(fs) ((fs)->blks_per_cluster * (fs)->blk_size)

/* predikat, zda je dany logicky cluster je reprezentovan primym odkazem, nebo ne */
#define DIRECT_LINK(log_cluster)	((log_cluster) < DIRECT_LINKS)

/* predikat, zda je dany logicky cluster je reprezentovan neprimym odkazem, nebo ne */
#define INDIRECT_LINK(fs, log_cluster)	(((log_cluster) >= DIRECT_LINKS) && ((log_cluster) < (DIRECT_LINKS + INDIRECT_LINKS(fs))))

/* predikat, zda dany fyzicky cluster je dira, nebo ne */
#define IS_EMPTY(cluster)	((cluster) == EMPTY_CLUSTER)

/* typy zaznamu v adresari */
#define	REC_INVALID	(0)
#define REC_FILE	(1)
#define REC_DIR         (2)

/* typ inode */
#define TYPE_FILE	(1)
#define TYPE_DIR	(2)

/* typ dotazu pro listovani adresare */
#define QUERY_INVALID_RECORD	(0)
#define QUERY_FILE              (1)
#define QUERY_ALL_FILES         (2)

/* zarovnani hodnoty na nasobky 4 */
#define ALIGN4(x)       (((x) + 3) & ~3)

typedef int32_t inode_id;       /* identifikator inode (musi byt znamenkove cislo, zaporne hodnoty signalizuji chybu) */
typedef int32_t cluster_id;     /* identifikator clusteru (musi byt znamenkove cislo, zaporne hodnoty signalizuji chybu)  */

/**
 * Struktura nesouci informace o souborovem systemu
 */
struct mufs {
	/* perzistentni cast dat */
	uint32_t blk_size;         /* velikost bloku (v bytech) */
	uint32_t superblk_blks;    /* velikost superbloku v blocich */
	uint32_t blks_per_cluster; /* pocet bloku v jednom clusteru */
	uint32_t inode_blks;       /* pocet bloku vyhrazenych pro inodes */
	uint32_t inode_bmap_blks;  /* pocet bloku vyhrazenych pro bitmapy s volnymi inodes */
	uint32_t data_bmap_blks;   /* pocet bloku vyhrazenych pro bitmapy s volnymi clustery */
	uint32_t data_cluster_cnt; /* pocet clusteru v datove casti */
	uint32_t first_data_blk;   /* index prvniho bloku s daty */

	/* in-memory cast dat */
	struct blk_dev *dev;
	struct inode *free_inodes; /** seznam nevyuzitych inodes z inode_bufferu */
};

/**
 * Struktura reprezentujici soubor
 */
struct inode {
	/* perzistentni cast dat */
	uint64_t size;                         /* velikost souboru v bytech */
	uint32_t type;                         /* typ inode */
	cluster_id direct_links[DIRECT_LINKS]; /* prime odkazy */
	cluster_id indirect_link;              /* neprime odkazy prvni urovne */
	uint32_t hardlinks;                    /* pocet linku na dany soubor */
	uint8_t reserved[20];                  /* volne misto pro dalsi pouziti */

	/* in-memory cast dat */
	uint32_t id;                           /* identifikator inode */
	uint32_t active_fds;                   /* pocet aktivnich file descriptoru */
	struct inode *next;                    /* odkaz na dalsi inode */
};

/**
 * Zaznam v adresari
 */
struct dir_entry {
	uint16_t rec_len;  /* velikost zaznamu vcetne hlavicky */
	uint8_t type;      /* typ zaznamu */
	uint8_t reserved;  /* vyhrazeno, zarovnani */
	inode_id inode;    /* inode souboru */
	char fname[255];   /* jmeno souboru vcetne \0 */
};

/**
 * Struktra pro vyhledavani zaznamu v adresari
 */
struct dir_query
{
	/* cast specifikujici, co se ma vyhledat */
	struct mufs *fs;
	struct inode *dir_inode;         /* inode adresare */
	uint8_t type;                    /* typ dotazu */
	union {
		uint32_t min_rec_size;   /* kdyz chceme prazdny zaznam, specifikujeme min. velikost */
		char *fname;             /* kdyz chceme konkretni soubor, specifikujeme jeho jmeno */ 
		void *data;              /* pouziva se pri vytvoreni dotazu */
	} constraint;

	/* atributy urcene k prohledavani adresare */
	uint64_t dir_offset;             /* offset nacteneho clusteru v ramci adresare */
	uint32_t cluster_offset;         /* offset zaznamu v ramci clusteru */
	unsigned char *cluster_buf;
};

/**
 * Struktura reprezentujici otevreny soubor
 */
struct mufs_fd {
	struct mufs *fs;
	struct inode *inode;
	uint64_t offset;      /* aktualni ukazatel v souboru */
	int flags;            /* dodatecne priznaky */
};

/**
 * Struktura reprezentujici otevreny adresar
 */
struct mufs_dird {
	struct dir_query query; /* dotaz, kterym je adresar prochazen */
	unsigned char *buf;     /* buffer s nactenou casti adresare */
};


/*
 * Nizkourovnove operace se souborovym systemem
 */
int mufs_format(struct blk_dev *dev, uint32_t blks_per_cluster, uint32_t inodes_cnt);
int mufs_open(struct mufs *fs, struct blk_dev *dev);
inode_id mufs_inode_create(struct mufs *fs, uint32_t type, struct inode *inode);
int mufs_write(struct mufs *fs, struct inode *inode, unsigned char *data, uint64_t offset, uint64_t size);
int mufs_read(struct mufs *fs, struct inode *inode, unsigned char *buf, uint64_t offset, uint64_t size);
int mufs_truncate(struct mufs *fs, struct inode *inode, uint64_t size);

/*
 * Operace s adresari
 */
int mufs_dir_create(struct mufs *fs, inode_id parent_dir, struct inode *dir_inode);
int mufs_dir_query(struct mufs *fs, struct dir_query *query, struct inode *dir_inode, unsigned char *cluster_buf, uint8_t query_type, void *constraint);
struct dir_entry *mufs_dir_next(struct dir_query *query);
int mufs_dir_link(struct mufs *fs, struct inode *dir_inode, inode_id inode, char *fname, int is_dir);
int mufs_dir_unlink(struct mufs *fs, struct inode *dir_inode, char *fname);

/*
 * Vysokourovnove rozhrani
 */
struct mufs_fd *mufs_file_open(struct mufs *fs, char *path, int flags);
int mufs_file_close(struct mufs_fd *fd);
uint64_t mufs_file_write(struct mufs_fd *fd, unsigned char *buf, uint64_t size);
uint64_t mufs_file_read(struct mufs_fd *fd, unsigned char *buf, uint64_t size);
uint64_t mufs_file_seek(struct mufs_fd *fd, int64_t offset, int whence);
int mufs_file_eof(struct mufs_fd *fd);
int mufs_file_unlink(struct mufs *fs, char *path);
int mufs_hardlink(struct mufs *fs, char *source_path, char *dest_path);

int mufs_mkdir(struct mufs *fs, char *path);
struct mufs_dird *mufs_dir_open(struct mufs *fs, char *path, int flags);
struct dir_entry *mufs_readdir(struct mufs_dird *dird);
void mufs_dir_close(struct mufs_dird *dird);
int mufs_rmdir(struct mufs *fs, char *path);

#endif
